#include <llog.h>

#include <stdio.h>
#include <string.h>
#include <time.h>

#if _OS_WIN

#elif _OS_LINUX

#else
#error "Not Supported Operator System"
#endif

#define LOG_COLOR(code) \033\133##code##m
#define LOG_COLOR_TRACE LOG_COLOR(94)
#define LOG_COLOR_DEBUG LOG_COLOR(36)
#define LOG_COLOR_INFO LOG_COLOR(32)
#define LOG_COLOR_WARN LOG_COLOR(33)
#define LOG_COLOR_ERROR LOG_COLOR(31)
#define LOG_COLOR_FATAL LOG_COLOR(35)

#define LOG_MAX_HEAD_BUF_SIZE 256

#define _TO_STR(str) #str
#define TO_STR(str) _TO_STR(str)

static const char* _log_level_strings[] = {
  "TRACE", "DEBUG", "INFO", "WARN", "ERROR", "FATAL"
};

static const char* _log_level_colors[] = {
	TO_STR(LOG_COLOR_TRACE), TO_STR(LOG_COLOR_DEBUG), TO_STR(LOG_COLOR_INFO),
	TO_STR(LOG_COLOR_WARN), TO_STR(LOG_COLOR_ERROR), TO_STR(LOG_COLOR_FATAL)
};

struct _fill_head_parm {
	int level;
	char* buf;
	size_t szbuf;
	log_mask_t mask;
	const char* file;
	const char* mode;
	const char* func;
	int line;

	struct tm* time;
};

size_t _log_fill_head_time_ymd(struct _fill_head_parm* parm) {
	log_mask_t mask = parm->mask;
	char* buf = parm->buf;
	size_t szbuf = parm->szbuf;
	struct tm* time = parm->time;
	if (LOG_MASK_TIME_YMD & mask) {
		if (LOG_MASK_YEAR & mask) {
			szbuf += strftime(buf + szbuf, LOG_MAX_HEAD_BUF_SIZE - szbuf, "%Y", time);
		}
		else {
			szbuf += sprintf(buf + szbuf, "Y");
		}
		if (LOG_MASK_MONTH & mask) {
			szbuf += strftime(buf + szbuf, LOG_MAX_HEAD_BUF_SIZE - szbuf, "-%m", time);
		}
		else {
			szbuf += sprintf(buf + szbuf, "-M");
		}
		if (LOG_MASK_DAY & mask) {
			szbuf += strftime(buf + szbuf, LOG_MAX_HEAD_BUF_SIZE - szbuf, "-%d", time);
		}
		else {
			szbuf += sprintf(buf + szbuf, "-D");
		}
	}
	return szbuf;
}

size_t _log_fill_head_time_hms(struct _fill_head_parm* parm) {
	log_mask_t mask = parm->mask;
	char* buf = parm->buf;
	size_t szbuf = parm->szbuf;
	struct tm* time = parm->time;
	if (LOG_MASK_TIME_HMS & mask) {
		if (LOG_MASK_HOUR & mask) {
			szbuf += strftime(buf + szbuf, LOG_MAX_HEAD_BUF_SIZE - szbuf, "%H", time);
		}
		else {
			szbuf += sprintf(buf + szbuf, "h");
		}
		if (LOG_MASK_MINUTE & mask) {
			szbuf += strftime(buf + szbuf, LOG_MAX_HEAD_BUF_SIZE - szbuf, ":%M", time);
		}
		else {
			szbuf += sprintf(buf + szbuf, ":m");
		}
		if (LOG_MASK_SECOND & mask) {
			szbuf += strftime(buf + szbuf, LOG_MAX_HEAD_BUF_SIZE - szbuf, ":%S", time);
		}
		else {
			szbuf += sprintf(buf + szbuf, ":s");
		}
	}
	return szbuf;
}



size_t _log_fill_head_detail(struct _fill_head_parm* parm) {
	log_mask_t mask = parm->mask;
	char* buf = parm->buf;
	size_t szbuf = parm->szbuf;
	int file_offset = 0;

	if (LOG_MASK_COLORS & mask) {
		szbuf += sprintf(buf + szbuf, "%s%-5s\033[0m \033[90m",
			_log_level_colors[parm->level],
			(LOG_MASK_LEVEL & mask) ? _log_level_strings[parm->level] : "nan");
	}
	else {
		szbuf += sprintf(buf + szbuf, "%-5s ",
			(LOG_MASK_LEVEL & mask) ? _log_level_strings[parm->level] : "nan");
	}
	if (LOG_FILE_SIMPLE & mask) {
		for (file_offset = strlen(parm->file);
			file_offset > -1; file_offset -= 1) {
			if (parm->file[file_offset] == '/' || parm->file[file_offset] == '\\') {
				file_offset += 1;
				break;
			}
		}
	}
	szbuf += sprintf(buf + szbuf, "%.128s:%.32s:%.32s:%d:",
		(LOG_MASK_FILE & mask) ? parm->file + file_offset : "file",
		(LOG_MASK_MODE & mask) ? parm->mode : "mode",
		(LOG_MASK_FUNC & mask) ? parm->func : "func",
		(LOG_MASK_LINE & mask) ? parm->line : 0);

	if (LOG_MASK_COLORS & mask) {
		szbuf += sprintf(buf + szbuf, "\033[0m");
	}
	return szbuf;
}

size_t log_fill_head_str(int level, char* log, size_t szlog,
	log_mask_t mask, const char* file, const char* mode, const char* func, int line) {
	size_t szbuf = 0;
	char buf[LOG_MAX_HEAD_BUF_SIZE] = { 0 };

	time_t t = time(NULL);
	struct tm* time = localtime(&t);
	struct _fill_head_parm parm = {
		.level = level,
		.buf = buf,
		.szbuf = 0,
		.mask = mask,
		.file = file,
		.mode = mode,
		.func = func,
		.line = line,

		.time = time
	};
	size_t(*callback[3])(struct _fill_head_parm* parm) = {
		_log_fill_head_time_ymd,
		_log_fill_head_time_hms,
		_log_fill_head_detail
	};
	for (int i = 0; i < sizeof(callback) / sizeof(callback[0]); i++) {
		szbuf += (*callback[i])(&parm);
		parm.buf[szbuf] = '|';
		parm.szbuf = szbuf + 1;
	}
	if (LOG_MASK_ENDLINE & mask) szbuf += sprintf(buf, "\n");
	buf[szbuf] = '\0';
	szbuf += 1;
	if (szlog <= szbuf) return 0;
	strncpy(log, buf, szbuf);
	return szbuf;
}
